/*
 *	JSCView
 *	(SwingOSC classes for SuperCollider)
 *
 *	Replacements for the basic (Cocoa) views.
 *	changes to v0.36 : - JSCDragView : added interpretDroppedStrings feature
 *	changes to v0.40 : - JSCCompositeView coordinates fixed
 *	changes to v0.42 : - JSCCompositeView is not opaque any more ; added opaque method
 *	changes to v0.43 : - added JSCMultiSliderView ; fixed JSCH/VLayoutView
 *
 *	Most changes to the cocoa classes are indicated by "JJJ begin" ... "JJJ end"
 *	blocks, although as the classes get more and more independant these comments
 *	tend to disappear.
 *
 *	For details, see JSCView.help.rtf and DeveloperInfo.rtf
 *
 *	@version		0.441, 16-Dec-06
 *	@author		SuperCollider Developers
 *	@author		Hanns Holger Rutz
 *
 *	@todo		should invoke custom dispose() methods on java gadgets
 *	@todo		missing classes : SCMultiSliderView subclasses
 *	@todo		missing functionality : resize_
 *	@todo		sliders : put action defers in osc responder in a collapse object
 *	@todo		need to check all default properties are readable
 *	@todo		should call revalidate() only if parent window is already showing
 *	@todo		lists: should check OSC package overflow and in that case send partial lists
 *	@todo		should use centralized OSCresponders / path responders instead of producing them for every gadget
 */
JSCView {  // abstract class
	classvar <>currentDrag, <>currentDragString;
	classvar <>globalKeyDownAction, <>globalKeyUpAction;

	var dataptr, <parent, <>action, <background;
	// JJJ begin
	var <mouseDownAction, <mouseUpAction, <mouseOverAction, <mouseMoveAction;
	// JJJ end
	var <>keyDownAction, <>keyUpAction, <>keyTyped;
	// JJJ begin
	var <beginDragAction,<>canReceiveDragHandler,<receiveDragHandler;
	// JJJ end
	var <>onClose;

	// JJJ begin
	var <server;
	var properties;
	var keyResp, dndResp, mouseResp, cmpResp;
	var <hasFocus = false;
	classvar fakeModifiers	= 0;	// reflects caps lock
	classvar unicodeMap;
	
	*initClass {
		unicodeMap = IdentityDictionary.new;
		// backspace and enter
		[ 8, 127, 10, 13,	
		// arrow keys and function keys
		  33, 63276, 34, 63277, 35, 63275, 36, 63273, 37, 63234, 38, 63232, 39, 63235, 40, 63233 ]
		.pairsDo({ arg key, value;
			unicodeMap.put( key, value );
		});
	}
	// JJJ end
	
	*new { arg parent, bounds, argServer, id;  // JJJ
		^super.new.init( parent, bounds, argServer, id );
	}
	
	// JJJ begin
	*prBasicNew {
		^super.new;
	}
	// JJJ end
	
	*viewClass { ^this }
	
	*paletteExample { arg parent, bounds;
		^this.new(parent, bounds);
	}
	
	init { arg argParent, argBounds, argServer, id;
		parent = argParent.asView; // actual view
		this.prInit( parent, argBounds.asRect, this.class.viewClass, argServer, id );
		argParent.add( this );// maybe window or viewadapter
	}
	
	asView { ^this }
	
	bounds {
		^this.getProperty(\bounds, Rect.new)
	}
	bounds_ { arg rect;
		this.setProperty(\bounds, rect)
	}
	
	visible {
		^this.getProperty(\visible)
	}
	visible_ { arg bool;
		this.setProperty(\visible, bool)
	}
	
	enabled {
		^this.getProperty(\enabled)
	}
	enabled_ { arg bool;
		this.setProperty(\enabled, bool)
	}
	
	canFocus {
		^this.getProperty(\canFocus)
	}
	canFocus_ { arg bool;
		this.setProperty(\canFocus, bool)
	}
	focus { arg flag=true;
		// JJJ begin
		if( flag, {
			server.sendMsg( '/method', this.id, \requestFocus );
		}, {
			"JSCView.focus( false ) : not yet implemented".error;
		});
		// JJJ end
//		_SCView_Focus
//		^this.primitiveFailed
	}

// JJJ ; turned into a field
//	hasFocus{
//		^false;
//		_SCView_Focus
//		^this.primitiveFailed
//	}

	id {
		^this.getProperty(\id)
	}
	id_ { arg id;
		this.setProperty(\id, id)
	}

	// JJJ begin
	opaque {
		^this.getProperty( \opaque );
	}
	
	opaque_ { arg bool;
		this.setProperty( \opaque, bool );
	}
	// JJJ end
	
	refresh {
		// JJJ begin
		server.sendBundle( nil,
			[ '/method', this.id, \revalidate ], [ '/method', this.id, \repaint ]);
		// JJJ end
//		_SCView_Refresh
//		^this.primitiveFailed
	}
	findByID { arg id;
		// JJJ begin
		"JSCView.findByID : not yet implemented".error;
		^nil;
		// JJJ end
//		_SCView_FindByID
//		^this.primitiveFailed
	}
	
	isClosed { ^dataptr.isNil }
	notClosed { ^dataptr.notNil }
	remove {
		if(dataptr.notNil,{
			parent.prRemoveChild(this);
			this.prRemove;
			this.prClose;
		},{
			"JSCView-remove : this view was already removed.".debug( this );
		});
	}
	/*
	resize behaviour in an SCCompositeView:
		1  2  3
		4  5  6
		7  8  9
		
		1 - fixed to left, fixed to top
		2 - horizontally elastic, fixed to top
		3 - fixed to right, fixed to top
		
		4 - fixed to left, vertically elastic
		5 - horizontally elastic, vertically elastic
		6 - fixed to right, vertically elastic
		
		7 - fixed to left, fixed to bottom
		8 - horizontally elastic, fixed to bottom
		9 - fixed to right, fixed to bottom
	*/
	resize {
		^this.getProperty(\resize)
	}
	resize_ { arg resize;
		// JJJ begin
		"JSCView.resize_ : not yet implemented".warn;
		// JJJ end
//		this.setProperty(\resize, resize)
	}
	
	background_ { arg color;
		background = color;
		this.setProperty(\background, color)
	}

	mouseDown{arg x, y, modifiers, buttonNumber, clickCount;
		mouseDownAction.value(this, x, y, modifiers, buttonNumber, clickCount);	
	}
	mouseUp{arg x, y, modifiers;
		mouseUpAction.value(this, x, y, modifiers);	
	}
	mouseMove{arg x, y, modifiers;
		mouseMoveAction.value(this, x, y, modifiers);	
	}	
	mouseOver{arg x, y;
		mouseOverAction.value(this, x, y);
	}
	
	// JJJ begin
	mouseDownAction_{ arg func;
		if( func.notNil && mouseResp.isNil, {this.prCreateMouseResponder; });
		mouseDownAction = func;
	}

	mouseUpAction_{ arg func;
		if( func.notNil && mouseResp.isNil, {this.prCreateMouseResponder; });
		mouseUpAction = func;
	}
	
	mouseOverAction_{ arg func;
		if( func.notNil && mouseResp.isNil, {this.prCreateMouseResponder; });
		mouseOverAction = func;
	}
	
	mouseMoveAction_{ arg func;
		if( func.notNil && mouseResp.isNil, {this.prCreateMouseResponder; });
		mouseMoveAction = func;
	}

	beginDragAction_ { arg func;
		if( func.notNil && dndResp.isNil, {this.prCreateDnDResponder; });
		beginDragAction = func;
	}
	
	receiveDragHandler_ { arg func;
		if( func.notNil && dndResp.isNil, {this.prCreateDnDResponder; });
		receiveDragHandler = func;
	}
	// JJJ end
	
	keyDown { arg char, modifiers, unicode,keycode;
		globalKeyDownAction.value(this, char, modifiers, unicode, keycode); 
		this.handleKeyDownBubbling(this, char, modifiers, unicode, keycode);
	}
	defaultKeyDownAction { ^nil }
	handleKeyDownBubbling { arg view, char, modifiers, unicode, keycode;
		var result;
		// nil from keyDownAction --> pass it on
		if (keyDownAction.isNil) {
			this.defaultKeyDownAction(char,modifiers,unicode,keycode);
			result = nil;
		}{
			result = keyDownAction.value(view, char, modifiers, unicode, keycode);
		};
		if(result.isNil) {  
			// call keydown action of parent view
			parent.handleKeyDownBubbling(view, char, modifiers, unicode, keycode);
		};
	}
	
	// sc.solar addition
	keyUp { arg char, modifiers, unicode,keycode; 
		this.keyTyped = char;
		// always call global keydown action first
		globalKeyUpAction.value(this, char, modifiers, unicode, keycode);
		this.handleKeyUpBubbling(this, char, modifiers, unicode, keycode);
	}
	defaultKeyUpAction { ^nil }
	handleKeyUpBubbling { arg view, char, modifiers,unicode,keycode;
		var result;
		// nil from keyDownAction --> pass it on
		if (keyUpAction.isNil) {
			this.defaultKeyUpAction(char,modifiers,unicode,keycode);
			result = nil;
		}{
			result = keyUpAction.value(view, char, modifiers, unicode, keycode);
		};
		if(result.isNil) {  
			// call keydown action of parent view
			parent.handleKeyUpBubbling(view, char, modifiers, unicode, keycode);
		};
	}

	beginDrag {
		currentDrag = if (beginDragAction.notNil) 
		{	
			beginDragAction.value(this)
		}{
			this.defaultGetDrag
		};
		currentDragString = currentDrag.asCompileString;
	}
	defaultGetDrag { ^nil }
	canReceiveDrag {
		^if(canReceiveDragHandler.notNil,{ canReceiveDragHandler.value(this) },{ this.defaultCanReceiveDrag })
	}
	defaultCanReceiveDrag { ^false }
	receiveDrag {
		if(receiveDragHandler.notNil,{ receiveDragHandler.value(this) },{ this.defaultReceiveDrag });
		currentDrag = currentDragString = nil;
	}

	// get the view parent tree up to the SCTopView
	getParents {
		var parents, view;
		view = this;
		parents = List.new;
		while({(view = view.parent).notNil},{ parents.add(view)});
		^parents
	}
	
	// JJJ begin
	prGetWindow {
		^parent.prGetWindow;
	}
	// JJJ end
	
	doAction {
		action.value(this);
	}
	
	
	properties {
		^#[\bounds, \visible, \enabled, \canFocus, \resize, \background,
				\minWidth,\maxWidth,\minHeight,\maxHeight, \opaque ] // JJJ: opaque
	}
	getPropertyList {
		^this.properties.collect({ arg name;
			[name, this.perform(name)]
		});
	}
	setPropertyList { arg list;
		list.do({ arg item;
			var name, value;
			#name, value = item;
			this.perform(name.asSetter, value);
		});
	}
	
	// private
	prInit { arg argParent, argBounds, argViewClass, argServer, argID;
		// JJJ begin
		server		= argServer ?? { argParent.server; };
		properties	= Dictionary.new;
		properties.put( \bounds, argBounds );
		properties.put( \visible, true );
		properties.put( \enabled, true );
		properties.put( \canFocus, true );
		properties.put( \resize, 1 );
		this.id		= argID ?? {server.nextNodeID; };
		dataptr		= this.id;

		^this.prSCViewNew;
		// JJJ end
//		_SCView_New
//		^this.primitiveFailed
	}
	
	// JJJ begin
	prClose { arg preMsg, postMsg;
		var bndl;
		
		// nil.remove is allowed
		keyResp.remove;
		dndResp.remove;
		mouseResp.remove;
		cmpResp.remove;
		
		bndl = List.new;
		bndl.addAll( preMsg );
		bndl.add([ '/method', "key" ++ this.id, \remove ]);
		bndl.add([ '/method', "cmp" ++ this.id, \remove ]);
		if( dndResp.notNil, {bndl.add([ '/method', "dnd" ++ this.id, \remove ]);});
		if( mouseResp.notNil, {bndl.add([ '/method', "mse" ++ this.id, \remove ]);});
		bndl.add([ '/free', "key" ++ this.id, "cmp" ++ this.id, this.id ] ++
			dndResp.notNil.if([ "dnd" ++ this.id ]) ++
			mouseResp.notNil.if([ "mse" ++ this.id ]) ++
			this.prIsInsideContainer.if([ "cn" ++ this.id ]);
		);
		bndl.addAll( postMsg );
//("tschuessi: "++bndl).postln;
		server.listSendBundle( nil, bndl );

		dataptr = nil;
		onClose.value(this);
	}
	// JJJ end

	// JJJ begin
	prSCViewNew {arg preMsg, postMsg;
		var bndl, argBounds;
		
		bndl = List.new;
		bndl.addAll( preMsg );
		argBounds = this.prBoundsToJava( this.bounds ).asSwingArg;
		bndl.add([ '/set', this.id, \bounds ] ++ argBounds ++ [ \font, '[', '/ref', \font, ']' ]);
		if( this.prIsInsideContainer, {
			bndl.add([ '/set', "cn" ++ this.id, \bounds ] ++ argBounds );
		});
		if( this.prNeedsTransferHandler, {
			this.prCreateDnDResponder( bndl );
		});
		// NOTE: for global key actions to be working, every view
		// has to create a key responder, even if it's not using it personally ;-(
		this.prCreateKeyResponder( bndl );
		this.prCreateCompResponder( bndl );
		bndl.addAll( postMsg );
		server.listSendBundle( nil, bndl );
	}
	// JJJ end
	
	// JJJ begin
	prCreateDnDResponder { arg bndl;
		var msg;
	
		if( dndResp.notNil, {
			"JSCView.prCreateDnDResponder : already created!".warn;
			^nil;
		});
		dndResp		= OSCresponderNode( server.addr, '/transfer', { arg time, resp, msg;
			var state;
			
			if( msg[1] === this.id, {
				state = msg[2].asSymbol;
				case { state === \export }
				{
					this.beginDrag;
					if( currentDrag.notNil, {
						server.listSendMsg([ '/set', "dnd" ++ this.id, \string ] ++ currentDrag.asString.asSwingArg );
					}, {
						server.sendMsg( '/set', "dnd" ++ this.id, \string, '[', '/ref', \null, ']' );
					});
				}
				{ state === \import }
				{
					if( msg[3].asSymbol === \string, {
						currentDrag = msg[4].asString;
						this.prImportDrag;	// compile it just as in cocoa sc
					});
					if( this.canReceiveDrag, {
						this.receiveDrag;
					});
				};
			});
		});
		dndResp.add;
		msg = [ '/local', "dnd" ++ this.id,
				'[', '/new', "de.sciss.swingosc.DummyTransferHandler", this.id, this.prGetDnDModifiers, ']' ];
		if( bndl.notNil, {
			bndl.add( msg );
		}, {
			server.sendBundle( nil, msg );
		});
	}
	// JJJ end
		

	// JJJ begin
	prCreateKeyResponder { arg bndl;
		var msg;
	
		if( keyResp.notNil, {
			"JSCView.prCreateKeyResponder : already created!".warn;
			^nil;
		});
		keyResp 		= OSCresponderNode( server.addr, '/key', { arg time, resp, msg;
			var char, state, propagate, unicode, modifiers, plusMod, keyCode;
		
			if( msg[1] === this.id, {
				state = msg[2].asSymbol;
				if( state !== \typed, {
					keyCode		= msg[3];
					unicode		= msg[4];
					modifiers		= msg[5];
					propagate		= unicode != 0xFFFF;
					plusMod		= fakeModifiers;
					if( propagate.not, {
						case { (keyCode >= 33) && (keyCode <= 40) } // arrow keys + page up/dn, home
						{
							plusMod   = plusMod |0x800000; // 0x900000;
							unicode   = unicodeMap.at( keyCode );
							propagate = true;
						}
						{ (keyCode >= 112) && (keyCode <= 123) } // F1 ... F12
						{
							plusMod   = plusMod |0x800000;
							unicode   = keyCode + 63124;
							propagate = true;
						}
						{ (keyCode >= 96) && (keyCode <= 111) } // numpad
						{
							plusMod   = plusMod |0x200000;
							propagate = true;
						}
						{ keyCode == 20 } // caps lock
						{
							// on/off is reflected thru keyPressed/keyReleased
							fakeModifiers = if( state === \pressed, 								fakeModifiers | 0x10000,
								fakeModifiers & 0x80FFF
							);
						};
					}, {
						unicode	= unicodeMap.atFail( keyCode, unicode );
					});
					if( propagate, {
						// java->cocoa ; this translates shift (1), ctrl (2), cmd (4), alt (8)
						modifiers		= ((modifiers & 3) << 17) |
									  ((modifiers & 4) << 18) |
									  ((modifiers & 8) << 16) | plusMod;
						char			= unicode.asAscii;
						if( state === \pressed, {
							{ this.keyDown( char, modifiers, unicode, keyCode );}.defer;
						}, { // "released
							{ this.keyUp( char, modifiers, unicode, keyCode );}.defer;
						});
					});
				});
			});
		});
		keyResp.add;
		msg = [ '/local', "key" ++ this.id, '[', '/new', "de.sciss.swingosc.KeyResponder", this.id, ']' ];
		if( bndl.notNil, {
			bndl.add( msg );
		}, {
			server.sendBundle( nil, msg );
		});
	}
	// JJJ end
		
	// JJJ begin
	prCreateCompResponder { arg bndl;
		var msg;
	
		if( cmpResp.notNil, {
			"JSCView.prCreateCompResponder : already created!".warn;
			^nil;
		});
		cmpResp 		= OSCresponderNode( server.addr, '/component', { arg time, resp, msg;
			var char, state;
		
			if( msg[1] === this.id, {
				state = msg[2].asSymbol;
				case {(state === \moved) || (state === \resized) }
				{
					properties.put( \bounds, this.prBoundsFromJava( Rect( msg[3], msg[4], msg[5], msg[6] )));
				}
				{ state === \gainedFocus }
				{
					hasFocus = true;
				}
				{ state === \lostFocus }
				{
					hasFocus = false;
				};
			});
		});
		cmpResp.add;
		msg = [ '/local', "cmp" ++ this.id, '[', '/new', "de.sciss.swingosc.ComponentResponder", this.id, ']' ];
		if( bndl.notNil, {
			bndl.add( msg );
		}, {
			server.sendBundle( nil, msg );
		});
	}
	// JJJ end

	// JJJ begin
	prCreateMouseResponder { arg bndl;
		var msg;
	
		if( mouseResp.notNil, {
			"JSCView.prCreateMouseResponder : already created!".warn;
			^nil;
		});
		mouseResp 		= OSCresponderNode( server.addr, '/mouse', { arg time, resp, msg;
			var state, x, y, modifiers, button, clickCount;
		
			if( msg[1] === this.id, {
				// [ '/mouse', id, state, x, y, modifiers, button, clickCount ]
				state 		= msg[2].asSymbol;
				x			= msg[3];
				y			= msg[4];
				modifiers		= msg[5];

				// java->cocoa ; this translates shift (1), ctrl (2), cmd (4), alt (8)
				modifiers		= ((modifiers & 3) << 17) |
							  ((modifiers & 4) << 18) |
							  ((modifiers & 8) << 16); // | plusMod;

				case { state === \pressed }
				{
					button		= msg[6];
					clickCount	= msg[7];
					{ this.mouseDown( x, y, modifiers, button, clickCount );}.defer;
				}
				{ state === \released }
				{
					{ this.mouseUp( x, y, modifiers );}.defer;
				}
				{ state === \moved }
				{
					{ this.mouseMove( x, y, modifiers );}.defer;
				}
				{ state === \dragged }
				{
					{ this.mouseMove( x, y, modifiers );}.defer;
				}
				{ state === \entered }
				{
					{ this.mouseOver( x, y );}.defer;
				};
			});
		});
		mouseResp.add;
		msg = [ '/local', "mse" ++ this.id, '[', '/new', "de.sciss.swingosc.MouseResponder", this.id, true, ']' ];
		if( bndl.notNil, {
			bndl.add( msg );
		}, {
			server.sendBundle( nil, msg );
		});
	}
	// JJJ end

	// JJJ begin
	prRemove {
//		_SCView_Remove
//		^this.primitiveFailed
	}
	// JJJ end
	
	// JJJ begin
	prIsInsideContainer {
		^false;	// default: no
	}
	
	prGetDnDModifiers {
		^2;	// default: control key
	}
	
	prNeedsTransferHandler {
		^false;
	}
	// JJJ end
	
	// "setProperty returns true if action needs to be called."
	setProperty { arg key, value;
		// JJJ begin
		var oldValue;
		
		oldValue	= properties.at( key );
		properties.put( key, value );
		this.prSendProperty( key, value );
		^(oldValue != value);
		// JJJ end
//		_SCView_SetProperty
//		^this.primitiveFailed
	}
	
	// JJJ begin
	prSendProperty { arg key, value;
		var bndl;

		key	= key.asSymbol;

		// fix keys
		switch( key,
			\id, {
				^nil; // not forwarded
			},
			\canFocus, {
				key = \focusable;
			},
			\bounds, {
				value = this.prBoundsToJava( value );
				if( this.prIsInsideContainer, {
					server.sendBundle( nil, [ '/set', this.id, key ] ++ value.asSwingArg,
								          [ '/set', "cn" ++ this.id, key ] ++ value.asSwingArg );
					^nil;
				});
			},
			\visible, {
				if( this.prIsInsideContainer, {
					server.sendMsg( '/set', "cn" ++ this.id, key, value.asSwingArg );
					^nil;
				});
			},
			\background, {
				if( value.isNil or: {value.alpha == 0.0 }, {
					if( this.opaque != false, {	// nil or false
						this.opaque_( false );
					});
				}, { if( value.alpha == 1.0, {
					if( this.opaque == false, {
						this.opaque_( true );
					});
				})});
			}
		);
		server.listSendMsg([ '/set', this.id, key ] ++ value.asSwingArg );
	}
	// JJJ end

	getProperty { arg key, value;
		// JJJ begin
		^properties.atFail( key, value );
		// JJJ end
//		_SCView_GetProperty
//		^this.primitiveFailed
	}	
	setPropertyWithAction { arg symbol, obj;
		// setting some properties may need to have the action called.
		if (this.setProperty(symbol, obj), {
			// setProperty returns true if action needs to be called.
			this.doAction;
		});
	}
	
	*importDrag { 
		// this is called when an NSString is the drag object
		// from outside of the SC app
		// we compile it to an SCObject.
		currentDragString = currentDrag;
		currentDrag = currentDrag.interpret;
	}
	
	// JJJ begin
	// this can be overridden
	prImportDrag {
		JSCView.importDrag;
	}
	// JJJ end

	// JJJ begin
	prBoundsToJava { arg rect;
		var pb;
		
		pb        = parent.bounds;
		^rect.moveBy( pb.left.neg, pb.top.neg );
	}
	// JJJ end

	// JJJ begin
	prBoundsFromJava { arg rect;
		var pb;
		
		pb        = parent.bounds;
		^rect.moveBy( pb.left, pb.top );
	}
	// JJJ end
}

JSCContainerView : JSCView { // abstract class
	var <children, <>decorator;
			
	add { arg child;
		var bndl;
	
		children = children.add(child);
		if (decorator.notNil, { decorator.place(child); });
		// JJJ begin
		bndl = List.new;
		bndl.add([ '/method', this.id, \add,
				'[', '/ref', child.prIsInsideContainer.if({ "cn" ++ child.id }, child.id ), ']' ]);
		if( this.prGetWindow.visible, {
			bndl.add([ '/method', this.id, \revalidate ]);
		});
		server.listSendBundle( nil, bndl );
		// JJJ end
	}
	
	removeAll {
		children.copy.do {|child| child.remove };
	}
	
	prRemoveChild { arg child;
		// JJJ begin
		children.remove(child);
		if( child.prIsInsideContainer, {
			server.sendMsg( '/method', this.id, \remove, '[', '/ref', "cn" ++ child.id, ']' );
		}, {
			server.sendMsg( '/method', this.id, \remove, '[', '/ref', child.id, ']' );
		});
		// JJJ end
		// ... decorator replace all
	}
	//bounds_  ... replace all

	prClose {
		super.prClose;
		children.do({ arg item; item.prClose });
	}

	// JJJ begin
	prSCViewNew { arg preMsg, postMsg;
		properties.put( \canFocus, false );
		^super.prSCViewNew( preMsg, postMsg );
	}
	// JJJ end
}

JSCCompositeView : JSCContainerView {
	// JJJ begin
	prSCViewNew {
		properties.put( \opaque, false );
		^super.prSCViewNew([
			[ '/set', '[', '/local', this.id, '[', '/new', "javax.swing.JPanel", ']', ']',
				\layout, '[', '/ref', \null, ']', \opaque, false ]
		]);
	}
	// JJJ end
}

JSCTopView : JSCContainerView {	// NOT subclass of JSCCompositeView
	// JJJ begin
	var window;
	// JJJ end

	// JJJ begin
	*new { arg window, bounds, argServer, id;  // JJJ
		^super.new( nil, bounds, argServer, id ).prInitTopView( window );
	}
	// JJJ end
	
	// JJJ begin
	prInitTopView { arg argWindow;
		window = argWindow;
	}
	// JJJ end

	// JJJ begin
	prGetWindow {
		^window;
	}
	// JJJ end

	// created by SCWindow
	handleKeyDownBubbling { arg view, char, modifiers, unicode, keycode;
		keyDownAction.value(view, char, modifiers, unicode, keycode);
	}
	handleKeyUpBubbling { arg view, char, modifiers, unicode, keycode;
		keyUpAction.value(view, char, modifiers, unicode, keycode);
	}

	//only in construction mode, handled internally
	canReceiveDrag { ^currentDrag.isKindOf(Class)}

	findWindow{
		JSCWindow.allWindows{|win|
			if(win.view == this){
				^win
			}
		}
	}
	
	defaultReceiveDrag{
		var win, view;
		win = this.findWindow;
		view = currentDrag.paletteExample(win, Rect(10,10,140,24));
		view.keyDownAction_({|view, char, modifiers, unicode, keycode|
			if(keycode == 51){
				view.remove;
			}
		});
	}

	// JJJ begin
	prSendProperty { arg key, value;
		var bndl;

		key	= key.asSymbol;

		switch( key,
			\bounds, { ^nil; }	// shouldn't resize content pane
		);
		^super.prSendProperty( key, value );
	}
	// JJJ end

	// JJJ begin
	prBoundsToJava { arg rect;
		^rect;
	}
	// JJJ end

	// JJJ begin
	prBoundsFromJava { arg rect;
		^rect;
	}
	// JJJ end
}

// JJJ : abstract!
JSCLayoutView : JSCContainerView {
	properties { ^super.properties ++ #[\spacing] }
	
	spacing {
		^this.getProperty(\spacing, 0)
	}
	spacing_ { arg distance;
		// JJJ begin
		"JSCLayoutView.spacing_ : not yet implemented".error;
		// JJJ end
		this.setProperty(\spacing, distance)
	}

	// JJJ begin
	prSCViewNew { arg preMsg, postMsg;
		properties.put( \spacing, 0 );
		^super.prSCViewNew( preMsg, postMsg );
	}
	// JJJ end
}

// XXX spacing doesn't work ; could use a SpringLayout in a future version
JSCHLayoutView : JSCLayoutView {

	// JJJ begin
	prSCViewNew {
		^super.prSCViewNew([
			[ '/set', '[', '/local', this.id, '[', '/new', "javax.swing.JPanel", ']', ']',
				\layout, '[', '/new', "javax.swing.BoxLayout", '[', '/ref', this.id, ']', 0, ']' ]
		]);
	}
	// JJJ end
}

// XXX spacing doesn't work ; could use a SpringLayout in a future version
JSCVLayoutView : JSCLayoutView {

	// JJJ begin
	prSCViewNew {
		^super.prSCViewNew([
			[ '/set', '[', '/local', this.id, '[', '/new', "javax.swing.JPanel", ']', ']',
				\layout, '[', '/new', "javax.swing.BoxLayout", '[', '/ref', this.id, ']', 1, ']' ]
		]);
	}
	// JJJ end
}


JSCControlView : JSCView { // abstract class
}

JSCSliderBase : JSCControlView {
	
	knobColor {
		^this.getProperty(\knobColor, Color.new)
	}
	knobColor_ { arg color;
		this.setProperty(\knobColor, color)
	}
	
	step_ { arg stepSize;
		this.setPropertyWithAction(\step, stepSize);
	}
	step {
		^this.getProperty(\step)
	}
	
	properties {
		^super.properties ++ #[\knobColor, \step]
	}

	// JJJ begin
	prSnap { arg val;
		if( this.step <= 0.0, {
			^val.clip( 0.0, 1.0 );
		}, {
			^(val.clip( 0.0, 1.0 ) / this.step).round * this.step;
		});
	}
	// JJJ end
}

JSCSlider : JSCSliderBase
{
	// JJJ begin
	var acResp, keyResp;	// OSCresponderNode for action listening
	var orientation;
	// JJJ end

	value {
		^this.getProperty(\value)
	}
	value_ { arg val;
		this.setProperty(\value, this.prSnap( val ));
	}	
	valueAction_ { arg val;
		this.setPropertyWithAction(\value, this.prSnap( val ));
	}	
	
	increment { ^this.valueAction = this.value + this.bounds.width.reciprocal }
	decrement { ^this.valueAction = this.value - this.bounds.width.reciprocal }
	
	defaultKeyDownAction { arg char, modifiers, unicode,keycode;
		// standard keydown
		if (char == $r, { this.valueAction = 1.0.rand; ^this });
		if (char == $n, { this.valueAction = 0.0; ^this });
		if (char == $x, { this.valueAction = 1.0; ^this });
		if (char == $c, { this.valueAction = 0.5; ^this });
		if (char == $], { this.increment; ^this });
		if (char == $[, { this.decrement; ^this });
		if (unicode == 16rF700, { this.increment; ^this });
		if (unicode == 16rF703, { this.increment; ^this });
		if (unicode == 16rF701, { this.decrement; ^this });
		if (unicode == 16rF702, { this.decrement; ^this });
		^nil		// bubble if it's an invalid key
	}
	
	defaultGetDrag { 
		^this.value
	}
	defaultCanReceiveDrag { 
		^currentDrag.isNumber 
	}
	defaultReceiveDrag {
		this.valueAction = currentDrag;
	}

	// JJJ begin
	prNeedsTransferHandler {
		^true;
	}
	// JJJ end

	thumbSize {
		^this.getProperty(\thumbSize, 12);
	}
	thumbSize_ { arg size;
		// JJJ begin
		"JSCSlider.thumbSize_ : not yet implemented".error;
		// JJJ end
		this.setProperty(\thumbSize, size);
	}
	
	properties {
		^super.properties ++ #[\thumbSize];
	}

	// JJJ begin
	prClose {
		acResp.remove;
		^super.prClose([[ '/method', "ac" ++ this.id, \remove ],
					   [ '/free', "ac" ++ this.id ]]);
	}
	// JJJ end

	// JJJ begin
	prSCViewNew {
		properties.put( \value, 0.0 );
		properties.put( \step, 0.0 );
		orientation = if( this.bounds.width > this.bounds.height, 0, 1 );
		acResp = OSCresponderNode( server.addr, '/action', { arg time, resp, msg;
			var newVal;
		
			case { msg[1] == this.id }
			{
				newVal = msg[4] / 0x40000000;
				if( newVal != this.value, {
					// don't call valueAction coz we'd create a loop
					properties.put( \value, newVal );
					{ this.doAction; }.defer;
				});
			};
		}).add;
		^super.prSCViewNew([
			[ '/local', this.id,
				'[', '/new', "de.sciss.swingosc.Slider", orientation, 0, 0x40000000, 0, ']',
				"ac" ++ this.id,
				'[', '/new', "de.sciss.swingosc.ActionResponder", this.id, \value, ']' ]
		]);
	}
	// JJJ end

	// JJJ begin
	prSendProperty { arg key, value;
		var bndl;

		key	= key.asSymbol;

		// fix keys
		case { key === \value }
		{
			key		= \valueNoAction;
			value	= value * 0x40000000;
		}
		{ key === \step }
		{
			value = max( 1, value * 0x40000000 ).asInteger;
			server.sendMsg( '/set', this.id, \snapToTicks, value != 0,
							\minorTickSpacing, value, \extent, value );
			^nil;
		}
		{ key === \bounds }
		{
			// SCButton paints focus outside bounds,
			// we have to paint it inside
			value 		= this.prBoundsToJava( value );
			if( if( value.width > value.height, 0, 1 ) != orientation, {
				orientation = 1 - orientation;
				server.sendMsg( '/set', this.id, \orientation, orientation );
			});
		};
		^super.prSendProperty( key, value );
	}
	// JJJ end
}

JSCKnob : JSCSlider
{
}

JSCRangeSlider : JSCSliderBase {
	// JJJ begin
	var acResp;	// OSCresponderNode for action listening
	// JJJ end

	*paletteExample { arg parent, bounds;
		var v;
		v = this.new(parent, bounds);
		v.lo = 0.2;
		v.hi = 0.7;
		^v
	}

	// JJJ begin
	step_ { arg stepSize;
		super.step_( stepSize );
		this.lo_( this.lo );
		this.hi_( this.hi );
	}
	// JJJ end
	
	lo {
		^this.getProperty(\lo)
	}
	lo_ { arg val;
		this.setProperty(\lo, this.prSnap( val ));	// JJJ
	}	
	activeLo_ { arg val;
		this.setPropertyWithAction(\lo, this.prSnap( val ));   // JJJ
	}	
	hi {
		^this.getProperty(\hi)
	}
	hi_ { arg val;
		this.setProperty(\hi, this.prSnap( val ));	// JJJ
	}	
	activeHi_ { arg val;
		this.setPropertyWithAction(\hi, this.prSnap( val ));   // JJJ
	}	
	range {
//		^this.getProperty(\range)
		^(this.getProperty( \hi ) - this.getProperty( \lo )).abs;   // JJJ
	}
	range_ { arg val;
//		this.setProperty(\range, val);
		this.setProperty( \hi, this.prSnap( this.getProperty(\lo) + val ));  // JJJ
	}	
	activeRange_ { arg val;
//		this.setPropertyWithAction(\range, val);
		// JJJ
		this.range_( val );
		this.doAction;
	}
	
	// JJJ begin
	setSpan {arg lo, hi;
		lo = this.prSnap( lo );
		hi = this.prSnap( hi );
		properties.put( \lo, lo );
		properties.put( \hi, hi );
		server.sendMsg( '/set', this.id, \knobX, min( lo, hi ), \knobWidth, abs( hi - lo ));
	}
	
	setSpanActive { arg lo, hi;
		this.setSpan( lo, hi );
		this.doAction;
	}
	// JJJ end

	properties {
		^super.properties ++ #[\lo, \hi]
	}
	
	increment { 
		var inc, val; 
		inc = this.bounds.width.reciprocal;
		val = this.hi + inc;
		if (val > 1, {
			inc = 1 - this.hi;
			val = 1;
		});
		// JJJ begin
		this.setSpanActive( this.lo + inc, val );
		// JJJ end
	}
	decrement { 
		var inc, val; 
		inc = this.bounds.width.reciprocal;
		val = this.lo - inc;
		if (val < 0, {
			inc = this.lo;
			val = 0;
		});
		// JJJ begin
		this.setSpanActive( val, this.hi - inc );
		// JJJ end
	}

	defaultKeyDownAction { arg char, modifiers, unicode;
		var a, b;
		// standard keydown
		if (char == $r, { 
			a = 1.0.rand; 
			b = 1.0.rand; 
			this.activeLo_(min(a, b));
			this.activeHi_(max(a, b));
			^this
		});
		// JJJ begin
		if (char == $n, { this.setSpanActive( 0.0, 0.0 ); ^this });
		if (char == $x, { this.setSpanActive( 1.0, 1.0 ); ^this });
		if (char == $c, { this.setSpanActive( 0.5, 0.5 ); ^this });
		if (char == $a, { this.setSpanActive( 0.0, 1.0 ); ^this });
		// JJJ end
		if (unicode == 16rF700, { this.increment; ^this });
		if (unicode == 16rF703, { this.increment; ^this });
		if (unicode == 16rF701, { this.decrement; ^this });
		if (unicode == 16rF702, { this.decrement; ^this });
		^nil		// bubble if it's an invalid key
	}
	defaultGetDrag { ^Point(this.lo, this.hi) }	
	defaultCanReceiveDrag {	
		^currentDrag.isKindOf(Point);	
	}
	defaultReceiveDrag {
		// changed to x,y instead of lo, hi
		this.lo = currentDrag.x;
		this.hi = currentDrag.y;
	}

	// JJJ begin
	prNeedsTransferHandler {
		^true;
	}
	// JJJ end

	// JJJ begin
	prClose {
		acResp.remove;
		^super.prClose([[ '/method', "ac" ++ this.id, \remove ],
					   [ '/free', "ac" ++ this.id ]]);
	}
	// JJJ end

	// JJJ begin
	prSCViewNew {
		properties.put( \lo, 0.0 );
		properties.put( \hi, 1.0 );
		properties.put( \step, 0.0 );
		acResp = OSCresponderNode( server.addr, '/action', { arg time, resp, msg;
			var newLo, newHi;
		
			case { msg[1] == this.id }
			{
				newLo	= msg[4];
				newHi 	= newLo + msg[6];
				if( (newLo != this.lo) || (newHi != this.hi), {
					// don't call valueAction coz we'd create a loop
					properties.put( \lo, newLo );
					properties.put( \hi, newHi );
					{ this.doAction; }.defer;
				});
			};
		}).add;
		^super.prSCViewNew([
			[ '/set', '[', '/local', this.id, '[', '/new', "de.sciss.swingosc.RangeSlider", ']', ']',
				\knobColor ] ++ Color.blue.asSwingArg,
			[ '/local', "ac" ++ this.id,
				'[', '/new', "de.sciss.swingosc.ActionResponder", this.id, '[', '/array', \knobX, \knobWidth, ']', ']' ]
		]);
	}
	// JJJ end

	// JJJ begin
	prSendProperty { arg key, value;
		key	= key.asSymbol;

		// fix keys
		case { key === \lo }
		{
			server.sendMsg( '/set', this.id, \knobX, min( value, this.hi ), \knobWidth, abs( value - this.hi ));
			^nil;		
		}
		{ key === \hi }
		{
			server.sendMsg( '/set', this.id, \knobX, min( value, this.lo ), \knobWidth, abs( value - this.lo ));
			^nil;		
		}
		{ key === \step }
		{
			key = \stepSize;
		};
		^super.prSendProperty( key, value );
	}
	// JJJ end

	// JJJ begin
	prBoundsToJava { arg rect;
		var pb;
		
		pb          = parent.bounds;
		// focus-border must be painted inside gadget bounds
		^rect.insetAll( -3, -3, -3, -3 ).moveBy( pb.left.neg, pb.top.neg );
	}

	prBoundsFromJava { arg rect;
		var pb;
		
		pb          = parent.bounds;
		^rect.moveBy( pb.left, pb.top ).insetAll( 3, 3, 3, 3 );
	}
	// JJJ end
}

JSC2DSlider : JSCSliderBase {
	// JJJ begin
	var acResp;	// OSCresponderNode for action listening
	// JJJ end

	// JJJ begin
	step_ { arg stepSize;
		super.step_( stepSize );
		this.x_( this.x );
		this.y_( this.y );
	}
	// JJJ end

	x {
		^this.getProperty(\x)
	}
	x_ { arg val;
		this.setProperty(\x, this.prSnap( val )); // JJJ
	}	
	activex_ { arg val;
		this.setPropertyWithAction(\x, this.prSnap( val )); // JJJ
	}	
	y {
		^this.getProperty(\y)
	}
	y_ { arg val;
		this.setProperty(\y, this.prSnap( val ));	// JJJ
	}	
	activey_ { arg val;
		this.setPropertyWithAction(\y, this.prSnap( val ));  // JJJ
	}	
	properties {
		^super.properties ++ #[\x, \y]
	}
	
	// JJJ begin
	setXY { arg x, y;
		x = this.prSnap( x );
		y = this.prSnap( y );
		properties.put( \x, x );
		properties.put( \y, y );
		server.sendMsg( '/set', this.id, \knobX, x, \knobY, y );
	}
	
	setXYActive { arg x, y;
		this.setXY( x, y );
		this.doAction;
	}
	// JJJ end

	incrementY { ^this.y = this.y + this.bounds.height.reciprocal }
	decrementY { ^this.y = this.y - this.bounds.height.reciprocal }
	incrementX { ^this.x = this.x + this.bounds.width.reciprocal }
	decrementX { ^this.x = this.x - this.bounds.width.reciprocal }

	defaultKeyDownAction { arg char, modifiers, unicode,keycode;
		// standard keydown
		// JJJ begin
		if (char == $r, { this.setXYActive( 1.0.rand, 1.0.rand ); ^this });
		if (char == $n, { this.setXYActive( 0.0, 0.0 ); ^this });
		if (char == $x, { this.setXYActive( 1.0, 1.0 ); ^this });
		if (char == $c, { this.setXYActive( 0.5, 0.5 ); ^this });
		// JJJ end
		if (unicode == 16rF700, { this.incrementY; this.doAction; ^this });
		if (unicode == 16rF703, { this.incrementX; this.doAction; ^this });
		if (unicode == 16rF701, { this.decrementY; this.doAction; ^this });
		if (unicode == 16rF702, { this.decrementX; this.doAction; ^this });
		^nil		// bubble if it's an invalid key
	}
	defaultGetDrag { 
		^Point(this.x, this.y)
	}
	defaultCanReceiveDrag {
		^currentDrag.isKindOf(Point);
	}
	defaultReceiveDrag {
		this.x = currentDrag.x;
		this.y = currentDrag.y;
	}

	// JJJ begin
	prNeedsTransferHandler {
		^true;
	}
	// JJJ end
	
	// JJJ begin
	prClose {
		acResp.remove;
		^super.prClose([[ '/method', "ac" ++ this.id, \remove ],
					   [ '/free', "ac" ++ this.id ]]);
	}
	// JJJ end

	// JJJ begin
	prSCViewNew {
		properties.put( \x, 0.0 );
		properties.put( \y, 0.0 );
		properties.put( \step, 0.0 );
		acResp = OSCresponderNode( server.addr, '/action', { arg time, resp, msg;
			var newX, newY;
		
			case { msg[1] == this.id }
			{
				newX = msg[4];
				newY = msg[6];
				if( (newX != this.x) || (newY != this.y), {
					// don't call valueAction coz we'd create a loop
					properties.put( \x, newX );
					properties.put( \y, newY );
					{ this.doAction; }.defer;
				});
			};
		}).add;
		^super.prSCViewNew([
			[ '/local', this.id,
				'[', '/new', "de.sciss.swingosc.Slider2D", ']',
				"ac" ++ this.id,
				'[', '/new', "de.sciss.swingosc.ActionResponder", this.id, '[', '/array', \knobX, \knobY, ']', ']' ]
		]);
	}
	// JJJ end

	// JJJ begin
	prSendProperty { arg key, value;
		var bndl;

		key	= key.asSymbol;

		// fix keys
		case { key === \x }
		{
			key = \knobX;
		}
		{ key === \y }
		{
			key = \knobY;
		}
		{ key === \step }
		{
			key = \stepSize;
		};
		^super.prSendProperty( key, value );
	}
	// JJJ end

	// JJJ begin
	prBoundsToJava { arg rect;
		var pb;
		
		pb = parent.bounds;
		^rect.insetAll( -3, -3, -3, -3 ).moveBy( pb.left.neg, pb.top.neg );
	}

	prBoundsFromJava { arg rect;
		var pb;
		
		pb          = parent.bounds;
		^rect.moveBy( pb.left, pb.top ).insetAll( 3, 3, 3, 3 );
	}
	// JJJ end
}

// JJJ : not yet working
JSC2DTabletSlider : JSC2DSlider {

//	var <>mouseDownAction,<>mouseUpAction;
	
	mouseDown { arg x,y,pressure,tiltx,tilty,deviceID,
			 buttonNumber,clickCount,absoluteZ,rotation;
		mouseDownAction.value(this,x,y,pressure,tiltx,tilty,deviceID, 
			buttonNumber,clickCount,absoluteZ,rotation);
	}
	mouseUp { arg x,y,pressure,tiltx,tilty,deviceID, 
			buttonNumber,clickCount,absoluteZ,rotation;
		mouseUpAction.value(this,x,y,pressure,tiltx,tilty,deviceID, 
			buttonNumber,clickCount,absoluteZ,rotation);
	}
	doAction { arg x,y,pressure,tiltx,tilty,deviceID, 
			buttonNumber,clickCount,absoluteZ,rotation;
		action.value(this,x,y,pressure,tiltx,tilty,deviceID, 
			buttonNumber,clickCount,absoluteZ,rotation);
	}
}

JSCButton : JSCControlView {
	var <font, <states;
	
	// JJJ begin
	var acResp;	// OSCresponderNode for action listening
	// JJJ end

	*paletteExample { arg parent, bounds;
		var v;
		v = this.new(parent, bounds);
		v.states = [
			["Push", Color.black, Color.red],
			["Pop", Color.white, Color.blue]];
		^v
	}
	
	value {
		^this.getProperty(\value)
	}
	value_ { arg val;
		this.setProperty(\value, this.prFixValue( val ));
	}	
	valueAction_ { arg val;
		this.setPropertyWithAction(\value, this.prFixValue( val ));
	}	

	// JJJ begin
	prFixValue { arg val;
		val = val.asInteger;
		// clip() would be better but SCButton resets to zero always
		if( (val < 0) || (val >= states.size), {
			val = 0;
		});
		^val;
	}
	// JJJ end
	
	doAction { arg modifiers;
		action.value(this, modifiers);
	}
	
	defaultKeyDownAction { arg char, modifiers, unicode;
// JJJ handled automatically by javax.swing.AbstractButton
//		if (char == $ , { this.valueAction = this.value + 1; ^this });
		if (char == $\r, { this.valueAction = this.value + 1; ^this });
		if (char == $\n, { this.valueAction = this.value + 1; ^this });
		if (char == 3.asAscii, { this.valueAction = this.value + 1; ^this });
		^nil		// bubble if it's an invalid key
	}

	font_ { arg argFont;
		font = argFont;
		this.setProperty(\font, font)
	}

	states_ { arg array;
		states = array;
		this.setProperty(\states, states);
	}
	
	properties {
		^super.properties ++ #[\value, \font, \states]
	}
	
	defaultGetDrag { 
		^this.value
	}
	defaultCanReceiveDrag {
		^currentDrag.isNumber or: { currentDrag.isKindOf(Function) };
	}
	defaultReceiveDrag {
		if (currentDrag.isNumber) {
			this.valueAction = currentDrag;
		}{
			this.action = currentDrag;
		};
	}

	// JJJ begin
	prNeedsTransferHandler {
		^true;
	}
	// JJJ end

	// JJJ begin
	prClose {
		acResp.remove;
		^super.prClose([[ '/method', "ac" ++ this.id, \remove ],
					   [ '/free', "ac" ++ this.id ]]);
	}
	// JJJ end

	// JJJ begin
	prSCViewNew {
		properties.put( \value, 0 );
		acResp = OSCresponderNode( server.addr, '/action', { arg time, resp, msg;
			case { msg[1] == this.id }
			{
				// don't call valueAction coz we'd create a loop
				properties.put( \value, msg[4] );
				{ this.doAction; }.defer;
			};
		}).add;
		^super.prSCViewNew([
//			[ '/method', [ '/local', this.id, [ '/new', "de.sciss.swingosc.MultiStateButton" ]],
//				"setInputMap", 0, [ '/ref', \null ]],	// kill default spacebar binding
			[ '/local', this.id, '[', '/new', "de.sciss.gui.MultiStateButton", ']',
				"ac" ++ this.id,
				'[', '/new', "de.sciss.swingosc.ActionResponder", this.id, \selectedIndex, ']' ]
		]);
	}
	// JJJ end

	// JJJ begin
	prSendProperty { arg key, value;
		var bndl, msg;

		key	= key.asSymbol;

		// fix keys
		case { key === \value }
		{
			key = \selectedIndex;
		}
		{ key === \states }
		{
			bndl	= List.new;
			bndl.add([ '/method', this.id, \removeAllItems ]);
			value.do({ arg state;
				msg = List.newUsing([ '/method', this.id, \addItem ]);
				msg.addAll( state[0].asSwingArg );
				msg.addAll( state[1].asSwingArg );
				msg.addAll( state[2].asSwingArg );
				bndl.add( msg.asArray );
			});
//("Sending '"++bndl++"'").postln;
			server.listSendBundle( nil, bndl );
			^nil;
		};
		^super.prSendProperty( key, value );
	}
	// JJJ end
	
	// JJJ begin
	prBoundsToJava { arg rect;
		var pb;
		
		pb = parent.bounds;
		^rect.insetAll( -3, -3, -3, -3 ).moveBy( pb.left.neg, pb.top.neg );
	}

	prBoundsFromJava { arg rect;
		var pb;
		
		pb          = parent.bounds;
		^rect.moveBy( pb.left, pb.top ).insetAll( 3, 3, 3, 3 );
	}
	// JJJ end
}


JSCPopUpMenu : JSCControlView {
	var <font, <items;
	
	// JJJ begin
	var acResp;	// OSCresponderNode for action listening
	// JJJ end

	*paletteExample { arg parent, bounds;
		var v;
		v = this.new(parent, bounds);
		v.items = #["linear","exponential","sine","welch","squared","cubed"];
		^v
	}
		
	value {
		^this.getProperty(\value)
	}
	value_ { arg val;
		this.setProperty(\value, this.prFixValue( val )); // JJJ
	}	
	valueAction_ { arg val;
		this.setPropertyWithAction(\value, this.prFixValue( val )); // JJJ
	}
	
	// JJJ begin
	prFixValue { arg val;
		^val.clip( 0, items.size - 1 );
	}
	// JJJ end

	defaultKeyDownAction { arg char, modifiers, unicode;
// JJJ used by lnf
//		if (char == $ , { this.valueAction = this.value + 1; ^this });
//		if (char == $\r, { this.valueAction = this.value + 1; ^this });
//		if (char == $\n, { this.valueAction = this.value + 1; ^this });
		if (char == 3.asAscii, { this.valueAction = this.value + 1; ^this });
//		if (unicode == 16rF700, { this.valueAction = this.value + 1; ^this });
		if (unicode == 16rF703, { this.valueAction = this.value + 1; ^this });
//		if (unicode == 16rF701, { this.valueAction = this.value - 1; ^this });
		if (unicode == 16rF702, { this.valueAction = this.value - 1; ^this });
		^nil		// bubble if it's an invalid key
	}
	font_ { arg argFont;
		font = argFont;
		this.setProperty(\font, font)
	}
	items_ { arg array;
		items = array;
		this.setProperty(\items, items);
	}
	stringColor {
		^this.getProperty(\stringColor, Color.new)
	}
	stringColor_ { arg color;
		this.setProperty(\stringColor, color)
	}
	
	properties {
		^super.properties ++ #[\value, \font, \items, \stringColor]
	}

	defaultGetDrag { 
		^this.value
	}
	defaultCanReceiveDrag {
		^currentDrag.isNumber;
	}
	defaultReceiveDrag {
		this.valueAction = currentDrag;
	}

	// JJJ begin
	prNeedsTransferHandler {
		^true;
	}
	// JJJ end

	// JJJ begin
	prClose {
		acResp.remove;
		^super.prClose([[ '/method', "ac" ++ this.id, \remove ],
					   [ '/free', "ac" ++ this.id ]]);
	}
	// JJJ end

	// JJJ begin
	prSCViewNew {
		properties.put( \value, 0 );
		acResp = OSCresponderNode( server.addr, '/action', { arg time, resp, msg;
			var newVal;
			
			case { msg[1] == this.id }
			{
				newVal = msg[4];
				if( newVal != this.value, {
					// don't call valueAction coz we'd create a loop
					properties.put( \value, newVal );
					{ this.doAction; }.defer;
				});
			};
		}).add;
		^super.prSCViewNew([
			[ '/local', this.id, '[', '/new', "javax.swing.JComboBox", ']',
				"ac" ++ this.id,
				'[', '/new', "de.sciss.swingosc.ActionResponder", this.id, \selectedIndex, ']' ]
		]);
	}
	// JJJ end
	
	// JJJ begin
	prSendProperty { arg key, value;
		var bndl;

		key	= key.asSymbol;

		// fix keys
		case { key === \value }
		{
			key = \selectedIndex;
		}
		{ key === \items }
		{
			bndl	= List.new;
			bndl.add([ '/method', this.id, \removeAllItems ]);
			value.do({ arg item;
				bndl.add([ '/method', this.id, \addItem ] ++ item.asSwingArg );			});
			server.listSendBundle( nil, bndl );
			^nil;
		};
		^super.prSendProperty( key, value );
	}
	// JJJ end

	// JJJ begin
	prBoundsToJava { arg rect;
		var pb;
		
		pb = parent.bounds;
		if( rect.height < 26, {
			rect			= Rect( rect.left - pb.left, rect.top - ((26 - rect.height) >> 1) - pb.top,
							   rect.width, 26 );
		}, {
			rect	= rect.moveBy( pb.left.neg, pb.top.neg );
		});
		^rect;
	}

// what shall we do ...
//	prBoundsFromJava { arg rect;
//		^rect;
//	}
	// JJJ end
}



JSCStaticTextBase : JSCView {
	var <string, <font, <object, <>setBoth=true;
	
	font_ { arg argFont;
		font = argFont;
		this.setProperty(\font, font)
	}
	
	string_ { arg argString;
		string = argString.asString;
		this.setProperty(\string, string)
	}
	align_ { arg align;
		this.setProperty(\align, align)
	}
	
	stringColor {
		^this.getProperty(\stringColor, Color.new)
	}
	stringColor_ { arg color;
		this.setProperty(\stringColor, color)
	}

	object_ { arg obj;
		object = obj;
		if (setBoth) { this.string = object.asString(80); };
	}
	
	properties {
		^super.properties ++ #[\string, \font, \stringColor]
	}

	// JJJ begin
	prSendProperty { arg key, value;
		var bndl;

		key	= key.asSymbol;

		// fix keys
		case { key === \stringColor }
		{
			key = \foreground;
		}
		{ key === \align }
		{
			key = \horizontalAlignment;
			case { value === \left }
			{
				value = 2;
			}
			{ value === \center }
			{
				value = 0;
			}
			{ value === \right }
			{
				value = 4;
			}
			// undocumented cocoa feature : -1 = left, 0 = center, 1 = right
			{ value.isKindOf( SimpleNumber )}
			{
				value = switch( value.sign, -1, 2, 0, 0, 1, 4 );
			};
		}
		{ key === \string }
		{
			key = \text;
//			value = value.asSwingArg;
// funktioniert nicht, weil BoundedRangeModel offensichtlich nochmal durch setText veraendert wird
//			server.sendBundle( nil,
//				[ '/set', this.id, \text, value ],	// make sure the text beginning is shown
//				[ "/methodr", [ '/method', this.id, \getHorizontalVisibility ], \setValue, 0 ]
//			);
//			^nil;
		};
		^super.prSendProperty( key, value );
	}
	// JJJ end
}

JSCStaticText : JSCStaticTextBase {
	*paletteExample { arg parent, bounds;
		var v;
		v = this.new(parent, bounds);
		v.string = "The lazy brown fox";
		^v
	}

	// JJJ begin
	prSCViewNew {
		properties.put( \canFocus, false );
		^super.prSCViewNew([
			[ '/local', this.id, '[', '/new', "de.sciss.swingosc.Label", ']' ]
		]);
	}
	// JJJ end
}

JSCListView : JSCControlView {
	var <font, <items, <>enterKeyAction;
	
	// JJJ begin
	var acResp;	// listens to list selection changes
	// JJJ end
	
	*paletteExample { arg parent, bounds;
		var v;
		v = this.new(parent, bounds);
		v.items = #["linear","exponential","sine","welch","squared","cubed"];
		^v
	}
	
	item {
		^items[this.value]
	}
	value {
		^this.getProperty(\value)
	}
	value_ { arg val;
		this.setProperty(\value, this.prFixValue( val ));
	}	
	valueAction_ { arg val;
		this.setPropertyWithAction(\value, this.prFixValue( val ));
	}	

	// JJJ begin
	prFixValue { arg val;
		val = val.asInteger;
		if( (val < 0) || (val >= items.size), {
			val = 0;
		});
		^val;
	}
	// JJJ end

	defaultKeyDownAction { arg char, modifiers, unicode;
		var index;
		if (char == $ , { this.valueAction = this.value + 1; ^this });
		if (char == $\r, { this.enterKeyAction.value(this); ^this });
		if (char == $\n, { this.enterKeyAction.value(this); ^this });
		if (char == 3.asAscii, { this.enterKeyAction.value(this); ^this });
// JJJ automatically handled by lnf
//		if (unicode == 16rF700, { this.valueAction = this.value - 1; ^this });
		if (unicode == 16rF703, { this.valueAction = this.value + 1; ^this });
//		if (unicode == 16rF701, { this.valueAction = this.value + 1; ^this });
		if (unicode == 16rF702, { this.valueAction = this.value - 1; ^this });
		if (char.isAlpha, {
			char = char.toUpper;
			index = items.detectIndex({|item| item.asString.at(0).toUpper >= char });
			if (index.notNil, {
				this.valueAction = index
			});
			^this
		});
		^nil		// bubble if it's an invalid key
	}
	font_ { arg argFont;
		font = argFont;
		this.setProperty(\font, font)
	}
	items_ { arg array;
		items = array;
		this.setProperty(\items, items);
	}
	stringColor {
		^this.getProperty(\stringColor, Color.new)
	}
	stringColor_ { arg color;
		this.setProperty(\stringColor, color)
	}
	
	selectedStringColor {
		^this.getProperty(\selectedStringColor, Color.new)
	}
	selectedStringColor_ { arg color;
		this.setProperty(\selectedStringColor, color)
	}
	
	hiliteColor {
		^this.getProperty(\hiliteColor, Color.new)
	}
	hiliteColor_ { arg color;
		this.setProperty(\hiliteColor, color)
	}
	
	properties {
		^super.properties ++ #[\value, \font, \items, \stringColor]
	}

	defaultGetDrag { 
		^this.value
	}
	defaultCanReceiveDrag {
		^currentDrag.isNumber;
	}
	defaultReceiveDrag {
		this.valueAction = currentDrag;
	}

	// JJJ begin
	prNeedsTransferHandler {
		^true;
	}
	// JJJ end

	// JJJ begin
	prClose {
		acResp.remove;
		^super.prClose([[ '/method', "ac" ++ this.id, \remove ],
					   [ '/free', "ac" ++ this.id ]]);
	}
	// JJJ end
	
	// JJJ begin
	prIsInsideContainer {
		^true;
	}
	// JJJ end

	// JJJ begin
	prSCViewNew {
		properties.put( \value, 0 );
		
		acResp = OSCresponderNode( server.addr, "/list", { arg time, resp, msg;
			var newVal;
		
			case { msg[1] == this.id }
			{
				newVal = this.prFixValue( msg[4] );
				// ignore -1 values which happen during dnd
				if( (msg[4] >= 0) && (newVal != this.value), {
//				if( newVal != this.value, {
					// don't call valueAction coz we'd create a loop
					properties.put( \value, newVal );
					{ this.doAction; }.defer;
				});
//				if( msg[6] == 0 ... )
			};
		}).add;
		^super.prSCViewNew([
			[ '/set', '[', '/local', this.id, '[', '/new', "javax.swing.JList", ']', ']',
				\selectionMode, 0 ],	// single selection only for compatibility
			[ '/local', "cn" ++ this.id, 	// bars : v=asNeeded, h=never
				'[', '/new', "javax.swing.JScrollPane", '[', '/ref', this.id, ']', 20, 31, ']',
				"ac" ++ this.id,
				'[', '/new', "de.sciss.swingosc.ListResponder", this.id,
					'[', '/array', \selectedIndex, ']', ']' ] // , \valueIsAdjusting
		]);
	}
	// JJJ end

	// JJJ begin
	prSendProperty { arg key, value;
		var bndl;

		key	= key.asSymbol;

		// fix keys
		case { key === \value }
		{
			key = \selectedIndex;
			server.sendBundle( nil,
				[ '/set', this.id, \selectedIndex, value.asSwingArg ],
				[ '/method', this.id, \ensureIndexIsVisible, value.asSwingArg ]
			);
			^nil;
		}
		{ key === \stringColor }
		{
			key = \foreground;
		}
		{ key === \selectedStringColor }
		{
			key = \selectionForeground;
		}
		{ key === \hiliteColor }
		{
			key = \selectionBackground;
		}
		{ key === \items }
		{
			server.listSendMsg([ '/set', this.id, \listData ] ++ value.asSwingArg ++
										[ \selectedIndex, this.value ]);
			^nil;
//		}
//		{ key === \bounds }
//		{
//			server.listSendMsg([ '/set', "cn" ++ this.id, key ] ++ this.prBoundsToJava( value ).asSwingArg );
//			^nil;
		};
		^super.prSendProperty( key, value );
	}
	// JJJ end
}

JSCDragView : JSCStaticTextBase {
	var <>interpretDroppedStrings = true;
	
	*paletteExample { arg parent, bounds;
		var v;
		v = this.new(parent, bounds);
		v.object = \something;
		^v
	}
	defaultGetDrag { 
		^object
	}

	// JJJ begin
	prNeedsTransferHandler {
		^true;
	}
	// JJJ end

	// JJJ begin
	prImportDrag {
		if( interpretDroppedStrings, {JSCView.importDrag; });
	}
	// JJJ end

	// JJJ begin
	prSCViewNew { arg preMsg, postMsg;
		properties.put( \canFocus, false );
		^super.prSCViewNew( preMsg, postMsg );
	}
	// JJJ end
}

JSCDragSource : JSCDragView {
	
	// JJJ begin
	prSCViewNew {
		^super.prSCViewNew([
			[ '/set', '[', '/local', this.id, '[', '/new', "de.sciss.swingosc.Label", ']', ']',
				\border, '[', '/method', "javax.swing.BorderFactory", \createRaisedBevelBorder, ']'
			]
		]);
	}
	// JJJ end

	// JJJ begin
	prGetDnDModifiers {
		^0;	// no modifiers needed
	}
	// JJJ end
}

JSCDragSink : JSCDragView {	
	defaultCanReceiveDrag { ^true;	}
	defaultReceiveDrag {
		this.object = currentDrag;
		this.doAction;
	}

	// JJJ begin
	prSCViewNew {
		^super.prSCViewNew([
			[ '/set', '[', '/local', this.id, '[', '/new', "de.sciss.swingosc.Label", ']', ']',
				\border, '[', '/method', "javax.swing.BorderFactory", \createLoweredBevelBorder, ']'
			]
		]);
	}
	// JJJ end

	// JJJ begin
	prGetDnDModifiers {
		^-1;	// don't allow it to be drag source
	}
	// JJJ end
}

JSCDragBoth : JSCDragView {		// JJJ not subclass of JSCDragSink
	// JJJ begin
	defaultCanReceiveDrag { ^true;	}
	defaultReceiveDrag {
		this.object = currentDrag;
		this.doAction;
	}
	// JJJ end

	defaultGetDrag { ^object }

	// JJJ begin
	prSCViewNew {
		^super.prSCViewNew([
			[ '/set', '[', '/local', this.id, '[', '/new', "de.sciss.swingosc.Label", ']', ']',
				\border, '[', '/method', "javax.swing.BorderFactory", \createCompoundBorder,
					'[', '/method', "javax.swing.BorderFactory", \createRaisedBevelBorder, ']',
					'[', '/method', "javax.swing.BorderFactory", \createLoweredBevelBorder, ']', ']'
			]
		]);
	}
	// JJJ end

	// JJJ begin
	prGetDnDModifiers {
		^0;	// no modifiers needed
	}
	// JJJ end
}


// JJJ : not yet working
JSCUserView : JSCView {
	var <>keyDownFunc, <drawFunc;
	// JJJ begin
	var <mouseBeginTrackFunc, <mouseTrackFunc, <mouseEndTrackFunc;
	// JJJ end
	
	// JJJ begin
	var penID			= nil;
	var isTracking	= false;
	// JJJ end

	draw {
		// JJJ begin
//		drawFunc.value(this)
		this.refresh;
		// JJJ end
	}
	
	// JJJ begin
	refresh {
		if( drawFunc.isNil, {
			server.sendMsg( '/method', this.id, \repaint );
		}, {
			JPen.protRefresh(drawFunc, this, server, penID, this.id );
		});
	}

	// it's unclear why mouseDown and mouseBeginTrack exist in parallel
	// ; we simply map the methods here
	mouseDown{ arg x, y, modifiers, buttonNumber, clickCount;
		this.mouseBeginTrack( x, y, modifiers );
		isTracking = true;
		^super.mouseDown( x, y, modifiers, buttonNumber, clickCount );
	}
	
	mouseUp{ arg x, y, modifiers;
		this.mouseEndTrack( x, y, modifiers );
		isTracking = false;
		^super.mouseUp( x, y, modifiers );
	}
	
	mouseMove{ arg x, y, modifiers;
		if( isTracking, {
			this.mouseTrack( x, y, modifiers );
		});
		^super.mouseMove( x, y, modifiers );	
	}
	
	mouseBeginTrackFunc_{ arg func;
		if( func.notNil && mouseResp.isNil, {this.prCreateMouseResponder; });
		mouseBeginTrackFunc = func;
	}

	mouseTrackFunc_{ arg func;
		if( func.notNil && mouseResp.isNil, {this.prCreateMouseResponder; });
		mouseTrackFunc = func;
	}

	mouseEndTrackFunc_{ arg func;
		if( func.notNil && mouseResp.isNil, {this.prCreateMouseResponder; });
		mouseEndTrackFunc = func;
	}
	// JJJ end
	
	mouseBeginTrack { arg x, y, modifiers; 
		mouseBeginTrackFunc.value(this, x, y, modifiers); 
	}
	mouseTrack { arg x, y, modifiers; 
		mouseTrackFunc.value(this, x, y, modifiers); 
	}
	mouseEndTrack { arg x, y, modifiers; 
		mouseEndTrackFunc.value(this, x, y, modifiers); 
	}
	keyDown { arg key, modifiers, unicode; 
		keyDownFunc.value(this, key, modifiers, unicode) 
	}

	// JJJ begin
	prSCViewNew {
		^super.prSCViewNew([
			[ '/local', this.id, '[', '/new', "javax.swing.JLabel", ']' ]
		]);
	}
	// JJJ end

	// JJJ begin
	drawFunc_ { arg func;
		if( drawFunc.isNil, {
			if( func.notNil, {
				penID	= server.nextNodeID;
				server.sendBundle( nil,
					[ '/local', penID, '[', '/new', "de.sciss.swingosc.Pen", '[', '/ref', this.id, ']', true, ']' ],
					[ '/method', this.id, \setIcon, '[', '/ref', penID, ']' ]
				);
				drawFunc = func;
				this.refresh;
			});
		}, {
			if( func.isNil, {
				server.sendBundle( nil,
					[ '/method', this.id, \setIcon, '[', '/ref', \null, ']' ],
					[ '/method', penID, \dispose ],
					[ '/free', penID ]
				);
				penID = nil;
				drawFunc = nil;
			}, {
				drawFunc = func;
				this.refresh;
			});
		});
	}
	// JJJ end

	// JJJ begin
	prClose {
		this.drawFunc_( nil );
		^super.prClose;
	}
	// JJJ end
}

// INCOMPLETE
JSCTextView : JSCView {
// JJJ
//	var <stringColor, <font, <editable;
//	var 	<autohidesScrollers, <hasHorizontalScroller, <hasVerticalScroller, <textBounds;
	var <stringColor, <font, <editable = true;
	var 	<autohidesScrollers = false, <hasHorizontalScroller = false, <hasVerticalScroller = false, <textBounds;
	var <usesTabToFocusNextView=true, <enterInterpretsSelection=true;
	mouseDown { arg clickPos;	
//		this.focus(true);
		mouseDownAction.value(this, clickPos);	
	}	
	
	string{
		^this.getProperty(\string);
	}

	string_{|str|
		^this.setString(str, -1);
	}
		
	selectedString {
		^this.getProperty(\selectedString);
	}
	
	selectedString_{|str|
		this.setProperty(\selectedString, str);	
	}
	
	selectionStart {
		^this.getProperty(\selectedRangeLocation);
	}
	
	selectionSize {
		^this.getProperty(\selectedRange);
	}	
	
	stringColor_ {arg color;
		stringColor = color;
		this.setStringColor(color,-1, 0);
	}
	
	setStringColor{arg color, rangeStart = -1, rangeSize = 0;
		this.setProperty(\setTextColor, [color,rangeStart, rangeSize]);
	}	
	font_ {arg afont;
		font = afont;
		this.setFont(font, -1, 0)
	}
	
	setFont{|	font, rangestart = -1, rangesize=0|
		this.setProperty(\setFont, [font, rangestart, rangesize]);
	}
	
	setString{|string, rangestart=0, rangesize=0|
		this.setProperty(\insertStringInRange, [string, rangestart, rangesize]);
	}
	
	editable_{|bool|
		editable = bool;
		this.setProperty(\setEditable, bool);
	}
	
// JJJ begin comment
//	enabled_{|bool|
//		this.editable(bool);
//	}
// JJJ end comment

	usesTabToFocusNextView_{|bool|
		usesTabToFocusNextView=bool;
		this.setProperty(\usesTabToFocusNextView, bool);
	}
	enterInterpretsSelection_{|bool|
		enterInterpretsSelection=bool;
		this.setProperty(\enterExecutesSelection, bool);
	}	
	autohidesScrollers_{|bool|
		autohidesScrollers = bool;
		this.setProperty(\setAutohidesScrollers, bool);
	}
	hasHorizontalScroller_{|bool|
		hasHorizontalScroller = bool;
		this.setProperty(\setHasHorizontalScroller, bool);
	}
	hasVerticalScroller_{|bool|
		hasVerticalScroller = bool;
		this.setProperty(\setHasVerticalScroller, bool);
	}	
	
	textBounds_{|rect|
		textBounds = rect;
		this.setProperty(\textBounds, rect);
	}		

	// JJJ begin
	prIsInsideContainer {
		^true;
	}
	// JJJ end

	// JJJ begin
	prSCViewNew {
//		properties.put( \value, 0 );
		
//		acResp = OSCresponderNode( server.addr, "/list", { arg time, resp, msg;
//			var newVal;
//		
//			case { msg[1] == this.id }
//			{
//				newVal = this.prFixValue( msg[4] );
//				// ignore -1 values which happen during dnd
//				if( (msg[4] >= 0) && (newVal != this.value), {
////				if( newVal != this.value, {
//					// don't call valueAction coz we'd create a loop
//					properties.put( \value, newVal );
//					{ this.doAction; }.defer;
//				});
////				if( msg[6] == 0 ... )
//			};
//		}).add;
		^super.prSCViewNew([
			[ '/local', this.id, '[', '/new', "javax.swing.JTextPane", ']',
				"cn" ++ this.id,				 	// bars : v=never, h=never
				'[', '/new', "javax.swing.JScrollPane", '[', '/ref', this.id, ']', 21, 31, ']',
				"dc" ++ this.id,
				'[', '/method', this.id, \getDocument, ']' ]
//			[ '/local', "ac" ++ this.id,
//				[ '/new', "de.sciss.swingosc.ListResponder", this.id,
//					[ '/array', \selectedIndex ]]] // , \valueIsAdjusting
		]);
	}
	// JJJ end

	// JJJ begin
	prSendProperty { arg key, value;
		var bndl, string, rangeStart, rangeSize;

		key	= key.asSymbol;

		// fix keys
		case { key === \insertStringInRange }
		{
			string		= value[ 0 ];
			rangeStart	= value[ 1 ];
			rangeSize		= value[ 2 ];
		
			bndl			= List.new;
			
			if( rangeStart == -1, {
				bndl.add([ '/method', "dc" ++ this.id, \remove, 0, '[', '/method', "dc" ++ this.id, \getLength, ']' ]);
				rangeStart = 0;
			}, { if( rangeSize > 0, {
				bndl.add([ '/method', "dc" ++ this.id, \remove, rangeStart, rangeSize ]);
			})});
			if( string.size > 0, { bndl.add([ '/method', "dc" ++ this.id, \insertString, rangeStart ] ++
				string.asSwingArg ++ nil.asSwingArg ); });
		
			if( bndl.notEmpty, {
				server.listSendBundle( nil, bndl );
			});
			^nil;
		}
//		{ }
//		{
//			\setCharacterAttributes, off, len, AttSet, false
//		}
		{ key === \setAutohidesScrollers or: {key === \setHasVerticalScroller or: { key === \setHasHorizontalScroller }} }
		{
//("hasH "++hasHorizontalScroller++"; hasV "++hasVerticalScroller++"; autoHide "++autohidesScrollers).postln;
			server.sendMsg( '/set', "cn" ++ this.id,
				\horizontalScrollBarPolicy, hasHorizontalScroller.if( autohidesScrollers.if( 30, 32 ), 31 ),
				\verticalScrollBarPolicy, hasVerticalScroller.if( autohidesScrollers.if( 20, 22 ), 21 ));
			^nil;
		}
		{ key === \setEditable }
		{
			key = \editable;
//		}
//		{ key == \bounds }
//		{
//			server.listSendMsg([ '/set', "cn" ++ this.id, key ] ++ this.prBoundsToJava( value ).asSwingArg );
//			^nil;
		};
		^super.prSendProperty( key, value );
	}
	// JJJ end
	
	// JJJ begin
	prClose {
//		acResp.remove;
//		^super.prClose([[ '/method', "ac" ++ this.id, \remove ],
//					   [ '/free', "dc" ++ this.id ]]);
		^super.prClose([[ '/free', "dc" ++ this.id ]]);
	}
	// JJJ end
}

JSCMultiSliderView : JSCView { 

	// JJJ begin
	var acResp;	// OSCresponderNode for action listening
	var vlResp;	// OSCresponderNode for value update listening
	// JJJ end

	var <>metaAction;
//	var <>mouseUpAction;
	var <>size ;
	var <gap;
	var <editable = true;
	var <elasticMode = 0;
	
// JJJ begin
//	draw {}
//	mouseBeginTrack { arg x, y, modifiers;}
//	mouseTrack { arg x, y, modifiers; 	}
//	mouseEndTrack { arg x, y, modifiers;
//		mouseUpAction.value(this);
//	}
// JJJ end
	
	properties {
		^super.properties ++ #[\value, \thumbSize, \fillColor, \strokeColor, \xOffset, \x, \y, \showIndex, \drawLines, \drawRects, \selectionSize, \startIndex, \referenceValues, \absoluteX, \isFilled, \step, \elasticResizeMode]; // JJJ not thumbWidth
	}	
	
	elasticMode_{arg mode;
		elasticMode =mode;
		this.setProperty(\elasticResizeMode, mode);
	}
	
	step_ { arg stepSize;
		this.setPropertyWithAction(\step, stepSize);
	}
	
	step {
		^this.getProperty(\step)
	}
	
	value { //returns array
		^this.getProperty(\value, Array.newClear(this.size))
	}
	value_ { arg val;
		this.size = val.size;
		this.setProperty(\value, val)
	}
	reference { //returns array
		^this.getProperty(\referenceValues, Array.newClear(this.size))
	}
	reference_ { arg val;
		//this.size = val.size;
		this.setProperty(\referenceValues, val)
	}
	index { //returns selected index
		^this.getProperty(\x)
	}
	index_ { arg inx;
		this.setProperty(\x, inx)
	}
	isFilled_ { arg abool;
		this.setProperty(\isFilled, abool);
	}
	xOffset_ { arg aval;
		this.setProperty(\xOffset, aval);
	}
	gap_ { arg inx;
		gap = inx;
		this.setProperty(\xOffset, inx)
	}
	
	selectionSize { 
		^this.getProperty(\selectionSize)
	}
	selectionSize_ { arg aval;
		this.setProperty(\selectionSize, aval)
	}
	currentvalue { //returns value of selected index
		^this.getProperty(\y)
	}
	fillColor_ { arg acolor;
		this.setProperty(\fillColor, acolor)
	}
	strokeColor_ { arg acolor;
		this.setProperty(\strokeColor, acolor)
	}
	colors_ { arg strokec, fillc;
		this.strokeColor_(strokec);
		this.fillColor_(fillc);
	}
	currentvalue_ { arg iny;
		this.setProperty(\y, iny)
	}
	showIndex_ { arg abool;
		this.setProperty(\showIndex, abool)
		}
	drawLines { arg abool;
		this.setProperty(\drawLines, abool)
	}
	drawLines_ { arg abool;
		this.drawLines(abool)
	}
	drawRects_ { arg abool;
		this.setProperty(\drawRects, abool)
	}
	readOnly_ { arg val;
		editable = val.not;
		this.setProperty(\readOnly, val);
	}
	editable_ { arg val;
		editable = val;
		this.setProperty(\readOnly, editable.not);
	}
	thumbSize_ { arg val;
		this.setProperty(\thumbSize, val)
	}
	indexThumbSize_ { arg val;
		this.setProperty(\indexThumbSize, val)
	}
	valueThumbSize_ { arg val;
		this.setProperty(\valueThumbSize, val)
	}
	indexIsHorizontal_ { arg val;
		this.setProperty(\isHorizontal,val);	
	}
	defaultCanReceiveDrag {	^true; }
	defaultReceiveDrag {
		if(currentDrag.at(0).isSequenceableCollection, { 
			this.value_(currentDrag.at(0));
			this.reference_(currentDrag.at(1));
		},{
			this.value_(currentDrag);
		});
	}	
	defaultGetDrag {
		var setsize, vals, rvals, outval;
		rvals = this.reference;
		vals = this.value;
		if(this.selectionSize > 1, {
			vals = vals.copyRange(this.index, this.selectionSize + this.index);});
		if(rvals.isNil, { 
			^vals 
		},{
		if(this.selectionSize > 1, {
			rvals = rvals.copyRange(this.index, this.selectionSize + this.index);});
			outval = outval.add(vals);
			outval = outval.add(rvals);
		});
		^outval
	}
		
	defaultKeyDownAction { arg key, modifiers, unicode;
		//modifiers.postln; 16rF702
		if (unicode == 16rF703, { this.index = this.index + 1; ^this });
		if (unicode == 16rF702, { this.index = this.index - 1; ^this });
		if (unicode == 16rF700, { this.gap = this.gap + 1; ^this });
		if (unicode == 16rF701, { this.gap = this.gap - 1; ^this });
		^nil		// bubble if it's an invalid key
	}
	
	doMetaAction{ 
		metaAction.value(this)
	} //on ctrl click

	// JJJ begin
	prNeedsTransferHandler {
		^true;
	}
	// JJJ end

	// JJJ begin
	prClose {
		vlResp.remove;
		acResp.remove;
		^super.prClose([[ '/method', "ac" ++ this.id, \remove ],
					   [ '/free', "ac" ++ this.id ]]);
	}
	// JJJ end

	// JJJ begin
	prSCViewNew {
		var initVal;
		initVal = 0 ! 8;
		properties.put( \value, initVal );
		vlResp = OSCresponderNode( server.addr, '/values', { arg time, resp, msg;
			var dirtyIndex, dirtySize, vals, selectedIndex;
			if( msg[ 1 ] === this.id, {
				vals			= properties[ \value ];
				dirtyIndex	= min( msg[ 2 ], vals.size );
				dirtySize		= min( msg[ 3 ], vals.size - dirtyIndex );
				
				dirtySize.do({ arg i;
					vals[ dirtyIndex + i ] = msg[ 4 + i ];
				});
				selectedIndex	= this.getProperty( \x, -1 );
//("selectedIndex = "++selectedIndex++"; vals = "++vals).inform;
				if( (selectedIndex >= dirtyIndex) and: { selectedIndex < (dirtyIndex + dirtySize) }, {
					properties.put( \y, vals[ selectedIndex ]);
				});
				{ this.doAction; }.defer;
			});
		}).add;
		acResp = OSCresponderNode( server.addr, '/action', { arg time, resp, msg;
			var dirtyIndex, dirtySize, vals, selectedIndex;
			if( msg[ 1 ] === this.id, {
				selectedIndex	= msg[ 4 ];
				properties.put( \x, selectedIndex );
				properties.put( \selectionSize, msg[ 6 ]);
				dirtyIndex	= msg[ 8 ];
				dirtySize		= msg[ 10 ];
//("dirtyIndex = "++dirtyIndex++"; dirtySize = "++dirtySize).inform;
				if( dirtySize > 0, {
					server.sendMsg( '/method', this.id, \sendValuesAndClear, this.id );
				}, {
					vals = properties[ \value ];
					properties.put( \y, vals[ selectedIndex ]);
//("y = "++properties[ \y ]).inform;
					{ this.doAction; }.defer;
				});
			});
		}).add;
		^super.prSCViewNew([
			[ '/local', this.id, '[', '/new', "de.sciss.swingosc.MultiSlider", ']',
				"ac" ++ this.id,
				'[', '/new', "de.sciss.swingosc.ActionResponder", this.id,
					'[', '/array', \selectedIndex, \selectionSize, \dirtyIndex, \dirtySize, ']', ']' ],
			[ '/set', this.id, \values ] ++ initVal.asSwingArg
		]);
	}
	// JJJ end

	// JJJ begin
	prSendProperty { arg key, value;
		var bndl, msg;

		key	= key.asSymbol;

		// fix keys
		case
		{ key === \value }
		{
			key = \values;
		}
		{ key === \x }
		{
			key = \selectedIndex;
		}
		{ key === \thumbSize }
		{
			key = \thumbWidth;
		}
		{ key === \indexThumbSize }
		{
			key = \thumbWidth;
		}
		{ key === \valueThumbSize }
		{
			key = \thumbHeight;
		}
		{ key === \isFilled }
		{
			key = \filled;
		}
		{ key === \isHorizontal }
		{
			key 		= \orientation;
			value	= if( value, 0, 1 );
		};
		^super.prSendProperty( key, value );
	}
	// JJJ end
	
	// JJJ begin
	prBoundsToJava { arg rect;
		var pb;
		
		pb = parent.bounds;
		^rect.insetAll( -3, -3, -3, -3 ).moveBy( pb.left.neg, pb.top.neg );
	}

	prBoundsFromJava { arg rect;
		var pb;
		
		pb          = parent.bounds;
		^rect.moveBy( pb.left, pb.top ).insetAll( 3, 3, 3, 3 );
	}
	// JJJ end
}